#ifndef __RESOURCE_nautilus_H__
#define __RESOURCE_nautilus_H__

#include <gio/gio.h>

G_GNUC_INTERNAL GResource *nautilus_get_resource (void);

G_GNUC_INTERNAL void nautilus_register_resource (void);
G_GNUC_INTERNAL void nautilus_unregister_resource (void);

#endif
