
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#ifndef NAUTILUS_ENUM_TYPES_H
#define NAUTILUS_ENUM_TYPES_H

#include <glib-object.h>

G_BEGIN_DECLS
/* Enumerations from "nautilus-enums.h" */
#define NAUTILUS_TYPE_GRID_ICON_SIZE (nautilus_grid_icon_size_get_type ())
GType nautilus_grid_icon_size_get_type (void);
#define NAUTILUS_TYPE_GRID_ZOOM_LEVEL (nautilus_grid_zoom_level_get_type ())
GType nautilus_grid_zoom_level_get_type (void);
#define NAUTILUS_TYPE_LIST_ICON_SIZE (nautilus_list_icon_size_get_type ())
GType nautilus_list_icon_size_get_type (void);
#define NAUTILUS_TYPE_LIST_ZOOM_LEVEL (nautilus_list_zoom_level_get_type ())
GType nautilus_list_zoom_level_get_type (void);
#define NAUTILUS_TYPE_FILE_ATTRIBUTES (nautilus_file_attributes_get_type ())
GType nautilus_file_attributes_get_type (void);
#define NAUTILUS_TYPE_MODE (nautilus_mode_get_type ())
GType nautilus_mode_get_type (void);
#define NAUTILUS_TYPE_OPEN_FLAGS (nautilus_open_flags_get_type ())
GType nautilus_open_flags_get_type (void);
#define NAUTILUS_TYPE_SEARCH_TIME_TYPE (nautilus_search_time_type_get_type ())
GType nautilus_search_time_type_get_type (void);
#define NAUTILUS_TYPE_SIDEBAR_ROW_TYPE (nautilus_sidebar_row_type_get_type ())
GType nautilus_sidebar_row_type_get_type (void);
#define NAUTILUS_TYPE_SIDEBAR_SECTION_TYPE (nautilus_sidebar_section_type_get_type ())
GType nautilus_sidebar_section_type_get_type (void);
/* Enumerations from "nautilus-search-popover.h" */
#define NAUTILUS_TYPE_SEARCH_FILTER (nautilus_search_filter_get_type ())
GType nautilus_search_filter_get_type (void);
G_END_DECLS

#endif /* NAUTILUS_ENUM_TYPES_H */

/* Generated data ends here */

