/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __NAUTILUS_GENERATED2_H__
#define __NAUTILUS_GENERATED2_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for org.gnome.Nautilus.FileOperations2 */

#define NAUTILUS_DBUS_TYPE_FILE_OPERATIONS2 (nautilus_dbus_file_operations2_get_type ())
#define NAUTILUS_DBUS_FILE_OPERATIONS2(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), NAUTILUS_DBUS_TYPE_FILE_OPERATIONS2, NautilusDBusFileOperations2))
#define NAUTILUS_DBUS_IS_FILE_OPERATIONS2(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), NAUTILUS_DBUS_TYPE_FILE_OPERATIONS2))
#define NAUTILUS_DBUS_FILE_OPERATIONS2_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), NAUTILUS_DBUS_TYPE_FILE_OPERATIONS2, NautilusDBusFileOperations2Iface))

struct _NautilusDBusFileOperations2;
typedef struct _NautilusDBusFileOperations2 NautilusDBusFileOperations2;
typedef struct _NautilusDBusFileOperations2Iface NautilusDBusFileOperations2Iface;

struct _NautilusDBusFileOperations2Iface
{
  GTypeInterface parent_iface;


  gboolean (*handle_copy_uris) (
    NautilusDBusFileOperations2 *object,
    GDBusMethodInvocation *invocation,
    const gchar *const *arg_sources,
    const gchar *arg_destination,
    GVariant *arg_platform_data);

  gboolean (*handle_create_folder) (
    NautilusDBusFileOperations2 *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_parent_uri,
    const gchar *arg_new_folder_name,
    GVariant *arg_platform_data);

  gboolean (*handle_delete_uris) (
    NautilusDBusFileOperations2 *object,
    GDBusMethodInvocation *invocation,
    const gchar *const *arg_uris,
    GVariant *arg_platform_data);

  gboolean (*handle_empty_trash) (
    NautilusDBusFileOperations2 *object,
    GDBusMethodInvocation *invocation,
    gboolean arg_ask_confirmation,
    GVariant *arg_platform_data);

  gboolean (*handle_move_uris) (
    NautilusDBusFileOperations2 *object,
    GDBusMethodInvocation *invocation,
    const gchar *const *arg_sources,
    const gchar *arg_destination,
    GVariant *arg_platform_data);

  gboolean (*handle_redo) (
    NautilusDBusFileOperations2 *object,
    GDBusMethodInvocation *invocation,
    GVariant *arg_platform_data);

  gboolean (*handle_rename_uri) (
    NautilusDBusFileOperations2 *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_uri,
    const gchar *arg_new_name,
    GVariant *arg_platform_data);

  gboolean (*handle_trash_uris) (
    NautilusDBusFileOperations2 *object,
    GDBusMethodInvocation *invocation,
    const gchar *const *arg_uris,
    GVariant *arg_platform_data);

  gboolean (*handle_undo) (
    NautilusDBusFileOperations2 *object,
    GDBusMethodInvocation *invocation,
    GVariant *arg_platform_data);

  gint  (*get_undo_status) (NautilusDBusFileOperations2 *object);

};

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (NautilusDBusFileOperations2, g_object_unref)
#endif

GType nautilus_dbus_file_operations2_get_type (void);

GDBusInterfaceInfo *nautilus_dbus_file_operations2_interface_info (void);
guint nautilus_dbus_file_operations2_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void nautilus_dbus_file_operations2_complete_copy_uris (
    NautilusDBusFileOperations2 *object,
    GDBusMethodInvocation *invocation);

void nautilus_dbus_file_operations2_complete_move_uris (
    NautilusDBusFileOperations2 *object,
    GDBusMethodInvocation *invocation);

void nautilus_dbus_file_operations2_complete_empty_trash (
    NautilusDBusFileOperations2 *object,
    GDBusMethodInvocation *invocation);

void nautilus_dbus_file_operations2_complete_trash_uris (
    NautilusDBusFileOperations2 *object,
    GDBusMethodInvocation *invocation);

void nautilus_dbus_file_operations2_complete_delete_uris (
    NautilusDBusFileOperations2 *object,
    GDBusMethodInvocation *invocation);

void nautilus_dbus_file_operations2_complete_create_folder (
    NautilusDBusFileOperations2 *object,
    GDBusMethodInvocation *invocation);

void nautilus_dbus_file_operations2_complete_rename_uri (
    NautilusDBusFileOperations2 *object,
    GDBusMethodInvocation *invocation);

void nautilus_dbus_file_operations2_complete_undo (
    NautilusDBusFileOperations2 *object,
    GDBusMethodInvocation *invocation);

void nautilus_dbus_file_operations2_complete_redo (
    NautilusDBusFileOperations2 *object,
    GDBusMethodInvocation *invocation);



/* D-Bus method calls: */
void nautilus_dbus_file_operations2_call_copy_uris (
    NautilusDBusFileOperations2 *proxy,
    const gchar *const *arg_sources,
    const gchar *arg_destination,
    GVariant *arg_platform_data,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean nautilus_dbus_file_operations2_call_copy_uris_finish (
    NautilusDBusFileOperations2 *proxy,
    GAsyncResult *res,
    GError **error);

gboolean nautilus_dbus_file_operations2_call_copy_uris_sync (
    NautilusDBusFileOperations2 *proxy,
    const gchar *const *arg_sources,
    const gchar *arg_destination,
    GVariant *arg_platform_data,
    GCancellable *cancellable,
    GError **error);

void nautilus_dbus_file_operations2_call_move_uris (
    NautilusDBusFileOperations2 *proxy,
    const gchar *const *arg_sources,
    const gchar *arg_destination,
    GVariant *arg_platform_data,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean nautilus_dbus_file_operations2_call_move_uris_finish (
    NautilusDBusFileOperations2 *proxy,
    GAsyncResult *res,
    GError **error);

gboolean nautilus_dbus_file_operations2_call_move_uris_sync (
    NautilusDBusFileOperations2 *proxy,
    const gchar *const *arg_sources,
    const gchar *arg_destination,
    GVariant *arg_platform_data,
    GCancellable *cancellable,
    GError **error);

void nautilus_dbus_file_operations2_call_empty_trash (
    NautilusDBusFileOperations2 *proxy,
    gboolean arg_ask_confirmation,
    GVariant *arg_platform_data,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean nautilus_dbus_file_operations2_call_empty_trash_finish (
    NautilusDBusFileOperations2 *proxy,
    GAsyncResult *res,
    GError **error);

gboolean nautilus_dbus_file_operations2_call_empty_trash_sync (
    NautilusDBusFileOperations2 *proxy,
    gboolean arg_ask_confirmation,
    GVariant *arg_platform_data,
    GCancellable *cancellable,
    GError **error);

void nautilus_dbus_file_operations2_call_trash_uris (
    NautilusDBusFileOperations2 *proxy,
    const gchar *const *arg_uris,
    GVariant *arg_platform_data,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean nautilus_dbus_file_operations2_call_trash_uris_finish (
    NautilusDBusFileOperations2 *proxy,
    GAsyncResult *res,
    GError **error);

gboolean nautilus_dbus_file_operations2_call_trash_uris_sync (
    NautilusDBusFileOperations2 *proxy,
    const gchar *const *arg_uris,
    GVariant *arg_platform_data,
    GCancellable *cancellable,
    GError **error);

void nautilus_dbus_file_operations2_call_delete_uris (
    NautilusDBusFileOperations2 *proxy,
    const gchar *const *arg_uris,
    GVariant *arg_platform_data,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean nautilus_dbus_file_operations2_call_delete_uris_finish (
    NautilusDBusFileOperations2 *proxy,
    GAsyncResult *res,
    GError **error);

gboolean nautilus_dbus_file_operations2_call_delete_uris_sync (
    NautilusDBusFileOperations2 *proxy,
    const gchar *const *arg_uris,
    GVariant *arg_platform_data,
    GCancellable *cancellable,
    GError **error);

void nautilus_dbus_file_operations2_call_create_folder (
    NautilusDBusFileOperations2 *proxy,
    const gchar *arg_parent_uri,
    const gchar *arg_new_folder_name,
    GVariant *arg_platform_data,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean nautilus_dbus_file_operations2_call_create_folder_finish (
    NautilusDBusFileOperations2 *proxy,
    GAsyncResult *res,
    GError **error);

gboolean nautilus_dbus_file_operations2_call_create_folder_sync (
    NautilusDBusFileOperations2 *proxy,
    const gchar *arg_parent_uri,
    const gchar *arg_new_folder_name,
    GVariant *arg_platform_data,
    GCancellable *cancellable,
    GError **error);

void nautilus_dbus_file_operations2_call_rename_uri (
    NautilusDBusFileOperations2 *proxy,
    const gchar *arg_uri,
    const gchar *arg_new_name,
    GVariant *arg_platform_data,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean nautilus_dbus_file_operations2_call_rename_uri_finish (
    NautilusDBusFileOperations2 *proxy,
    GAsyncResult *res,
    GError **error);

gboolean nautilus_dbus_file_operations2_call_rename_uri_sync (
    NautilusDBusFileOperations2 *proxy,
    const gchar *arg_uri,
    const gchar *arg_new_name,
    GVariant *arg_platform_data,
    GCancellable *cancellable,
    GError **error);

void nautilus_dbus_file_operations2_call_undo (
    NautilusDBusFileOperations2 *proxy,
    GVariant *arg_platform_data,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean nautilus_dbus_file_operations2_call_undo_finish (
    NautilusDBusFileOperations2 *proxy,
    GAsyncResult *res,
    GError **error);

gboolean nautilus_dbus_file_operations2_call_undo_sync (
    NautilusDBusFileOperations2 *proxy,
    GVariant *arg_platform_data,
    GCancellable *cancellable,
    GError **error);

void nautilus_dbus_file_operations2_call_redo (
    NautilusDBusFileOperations2 *proxy,
    GVariant *arg_platform_data,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean nautilus_dbus_file_operations2_call_redo_finish (
    NautilusDBusFileOperations2 *proxy,
    GAsyncResult *res,
    GError **error);

gboolean nautilus_dbus_file_operations2_call_redo_sync (
    NautilusDBusFileOperations2 *proxy,
    GVariant *arg_platform_data,
    GCancellable *cancellable,
    GError **error);



/* D-Bus property accessors: */
gint nautilus_dbus_file_operations2_get_undo_status (NautilusDBusFileOperations2 *object);
void nautilus_dbus_file_operations2_set_undo_status (NautilusDBusFileOperations2 *object, gint value);


/* ---- */

#define NAUTILUS_DBUS_TYPE_FILE_OPERATIONS2_PROXY (nautilus_dbus_file_operations2_proxy_get_type ())
#define NAUTILUS_DBUS_FILE_OPERATIONS2_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), NAUTILUS_DBUS_TYPE_FILE_OPERATIONS2_PROXY, NautilusDBusFileOperations2Proxy))
#define NAUTILUS_DBUS_FILE_OPERATIONS2_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), NAUTILUS_DBUS_TYPE_FILE_OPERATIONS2_PROXY, NautilusDBusFileOperations2ProxyClass))
#define NAUTILUS_DBUS_FILE_OPERATIONS2_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), NAUTILUS_DBUS_TYPE_FILE_OPERATIONS2_PROXY, NautilusDBusFileOperations2ProxyClass))
#define NAUTILUS_DBUS_IS_FILE_OPERATIONS2_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), NAUTILUS_DBUS_TYPE_FILE_OPERATIONS2_PROXY))
#define NAUTILUS_DBUS_IS_FILE_OPERATIONS2_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), NAUTILUS_DBUS_TYPE_FILE_OPERATIONS2_PROXY))

typedef struct _NautilusDBusFileOperations2Proxy NautilusDBusFileOperations2Proxy;
typedef struct _NautilusDBusFileOperations2ProxyClass NautilusDBusFileOperations2ProxyClass;
typedef struct _NautilusDBusFileOperations2ProxyPrivate NautilusDBusFileOperations2ProxyPrivate;

struct _NautilusDBusFileOperations2Proxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  NautilusDBusFileOperations2ProxyPrivate *priv;
};

struct _NautilusDBusFileOperations2ProxyClass
{
  GDBusProxyClass parent_class;
};

GType nautilus_dbus_file_operations2_proxy_get_type (void);

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (NautilusDBusFileOperations2Proxy, g_object_unref)
#endif

void nautilus_dbus_file_operations2_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
NautilusDBusFileOperations2 *nautilus_dbus_file_operations2_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
NautilusDBusFileOperations2 *nautilus_dbus_file_operations2_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void nautilus_dbus_file_operations2_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
NautilusDBusFileOperations2 *nautilus_dbus_file_operations2_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
NautilusDBusFileOperations2 *nautilus_dbus_file_operations2_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define NAUTILUS_DBUS_TYPE_FILE_OPERATIONS2_SKELETON (nautilus_dbus_file_operations2_skeleton_get_type ())
#define NAUTILUS_DBUS_FILE_OPERATIONS2_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), NAUTILUS_DBUS_TYPE_FILE_OPERATIONS2_SKELETON, NautilusDBusFileOperations2Skeleton))
#define NAUTILUS_DBUS_FILE_OPERATIONS2_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), NAUTILUS_DBUS_TYPE_FILE_OPERATIONS2_SKELETON, NautilusDBusFileOperations2SkeletonClass))
#define NAUTILUS_DBUS_FILE_OPERATIONS2_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), NAUTILUS_DBUS_TYPE_FILE_OPERATIONS2_SKELETON, NautilusDBusFileOperations2SkeletonClass))
#define NAUTILUS_DBUS_IS_FILE_OPERATIONS2_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), NAUTILUS_DBUS_TYPE_FILE_OPERATIONS2_SKELETON))
#define NAUTILUS_DBUS_IS_FILE_OPERATIONS2_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), NAUTILUS_DBUS_TYPE_FILE_OPERATIONS2_SKELETON))

typedef struct _NautilusDBusFileOperations2Skeleton NautilusDBusFileOperations2Skeleton;
typedef struct _NautilusDBusFileOperations2SkeletonClass NautilusDBusFileOperations2SkeletonClass;
typedef struct _NautilusDBusFileOperations2SkeletonPrivate NautilusDBusFileOperations2SkeletonPrivate;

struct _NautilusDBusFileOperations2Skeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  NautilusDBusFileOperations2SkeletonPrivate *priv;
};

struct _NautilusDBusFileOperations2SkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType nautilus_dbus_file_operations2_skeleton_get_type (void);

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (NautilusDBusFileOperations2Skeleton, g_object_unref)
#endif

NautilusDBusFileOperations2 *nautilus_dbus_file_operations2_skeleton_new (void);


G_END_DECLS

#endif /* __NAUTILUS_GENERATED2_H__ */
