/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __XDG_DESKTOP_PORTAL_DBUS_H__
#define __XDG_DESKTOP_PORTAL_DBUS_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for org.freedesktop.impl.portal.FileChooser */

#define XDP_IMPL_TYPE_FILE_CHOOSER (xdp_impl_file_chooser_get_type ())
#define XDP_IMPL_FILE_CHOOSER(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), XDP_IMPL_TYPE_FILE_CHOOSER, XdpImplFileChooser))
#define XDP_IMPL_IS_FILE_CHOOSER(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), XDP_IMPL_TYPE_FILE_CHOOSER))
#define XDP_IMPL_FILE_CHOOSER_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), XDP_IMPL_TYPE_FILE_CHOOSER, XdpImplFileChooserIface))

struct _XdpImplFileChooser;
typedef struct _XdpImplFileChooser XdpImplFileChooser;
typedef struct _XdpImplFileChooserIface XdpImplFileChooserIface;

struct _XdpImplFileChooserIface
{
  GTypeInterface parent_iface;

  gboolean (*handle_open_file) (
    XdpImplFileChooser *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_handle,
    const gchar *arg_app_id,
    const gchar *arg_parent_window,
    const gchar *arg_title,
    GVariant *arg_options);

  gboolean (*handle_save_file) (
    XdpImplFileChooser *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_handle,
    const gchar *arg_app_id,
    const gchar *arg_parent_window,
    const gchar *arg_title,
    GVariant *arg_options);

  gboolean (*handle_save_files) (
    XdpImplFileChooser *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_handle,
    const gchar *arg_app_id,
    const gchar *arg_parent_window,
    const gchar *arg_title,
    GVariant *arg_options);

};

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (XdpImplFileChooser, g_object_unref)
#endif

GType xdp_impl_file_chooser_get_type (void);

GDBusInterfaceInfo *xdp_impl_file_chooser_interface_info (void);
guint xdp_impl_file_chooser_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void xdp_impl_file_chooser_complete_open_file (
    XdpImplFileChooser *object,
    GDBusMethodInvocation *invocation,
    guint response,
    GVariant *results);

void xdp_impl_file_chooser_complete_save_file (
    XdpImplFileChooser *object,
    GDBusMethodInvocation *invocation,
    guint response,
    GVariant *results);

void xdp_impl_file_chooser_complete_save_files (
    XdpImplFileChooser *object,
    GDBusMethodInvocation *invocation,
    guint response,
    GVariant *results);



/* D-Bus method calls: */
void xdp_impl_file_chooser_call_open_file (
    XdpImplFileChooser *proxy,
    const gchar *arg_handle,
    const gchar *arg_app_id,
    const gchar *arg_parent_window,
    const gchar *arg_title,
    GVariant *arg_options,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean xdp_impl_file_chooser_call_open_file_finish (
    XdpImplFileChooser *proxy,
    guint* out_response,
    GVariant ** out_results,
    GAsyncResult *res,
    GError **error);

gboolean xdp_impl_file_chooser_call_open_file_sync (
    XdpImplFileChooser *proxy,
    const gchar *arg_handle,
    const gchar *arg_app_id,
    const gchar *arg_parent_window,
    const gchar *arg_title,
    GVariant *arg_options,
    guint* out_response,
    GVariant ** out_results,
    GCancellable *cancellable,
    GError **error);

void xdp_impl_file_chooser_call_save_file (
    XdpImplFileChooser *proxy,
    const gchar *arg_handle,
    const gchar *arg_app_id,
    const gchar *arg_parent_window,
    const gchar *arg_title,
    GVariant *arg_options,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean xdp_impl_file_chooser_call_save_file_finish (
    XdpImplFileChooser *proxy,
    guint* out_response,
    GVariant ** out_results,
    GAsyncResult *res,
    GError **error);

gboolean xdp_impl_file_chooser_call_save_file_sync (
    XdpImplFileChooser *proxy,
    const gchar *arg_handle,
    const gchar *arg_app_id,
    const gchar *arg_parent_window,
    const gchar *arg_title,
    GVariant *arg_options,
    guint* out_response,
    GVariant ** out_results,
    GCancellable *cancellable,
    GError **error);

void xdp_impl_file_chooser_call_save_files (
    XdpImplFileChooser *proxy,
    const gchar *arg_handle,
    const gchar *arg_app_id,
    const gchar *arg_parent_window,
    const gchar *arg_title,
    GVariant *arg_options,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean xdp_impl_file_chooser_call_save_files_finish (
    XdpImplFileChooser *proxy,
    guint* out_response,
    GVariant ** out_results,
    GAsyncResult *res,
    GError **error);

gboolean xdp_impl_file_chooser_call_save_files_sync (
    XdpImplFileChooser *proxy,
    const gchar *arg_handle,
    const gchar *arg_app_id,
    const gchar *arg_parent_window,
    const gchar *arg_title,
    GVariant *arg_options,
    guint* out_response,
    GVariant ** out_results,
    GCancellable *cancellable,
    GError **error);



/* ---- */

#define XDP_IMPL_TYPE_FILE_CHOOSER_PROXY (xdp_impl_file_chooser_proxy_get_type ())
#define XDP_IMPL_FILE_CHOOSER_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), XDP_IMPL_TYPE_FILE_CHOOSER_PROXY, XdpImplFileChooserProxy))
#define XDP_IMPL_FILE_CHOOSER_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), XDP_IMPL_TYPE_FILE_CHOOSER_PROXY, XdpImplFileChooserProxyClass))
#define XDP_IMPL_FILE_CHOOSER_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), XDP_IMPL_TYPE_FILE_CHOOSER_PROXY, XdpImplFileChooserProxyClass))
#define XDP_IMPL_IS_FILE_CHOOSER_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), XDP_IMPL_TYPE_FILE_CHOOSER_PROXY))
#define XDP_IMPL_IS_FILE_CHOOSER_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), XDP_IMPL_TYPE_FILE_CHOOSER_PROXY))

typedef struct _XdpImplFileChooserProxy XdpImplFileChooserProxy;
typedef struct _XdpImplFileChooserProxyClass XdpImplFileChooserProxyClass;
typedef struct _XdpImplFileChooserProxyPrivate XdpImplFileChooserProxyPrivate;

struct _XdpImplFileChooserProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  XdpImplFileChooserProxyPrivate *priv;
};

struct _XdpImplFileChooserProxyClass
{
  GDBusProxyClass parent_class;
};

GType xdp_impl_file_chooser_proxy_get_type (void);

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (XdpImplFileChooserProxy, g_object_unref)
#endif

void xdp_impl_file_chooser_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
XdpImplFileChooser *xdp_impl_file_chooser_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
XdpImplFileChooser *xdp_impl_file_chooser_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void xdp_impl_file_chooser_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
XdpImplFileChooser *xdp_impl_file_chooser_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
XdpImplFileChooser *xdp_impl_file_chooser_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define XDP_IMPL_TYPE_FILE_CHOOSER_SKELETON (xdp_impl_file_chooser_skeleton_get_type ())
#define XDP_IMPL_FILE_CHOOSER_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), XDP_IMPL_TYPE_FILE_CHOOSER_SKELETON, XdpImplFileChooserSkeleton))
#define XDP_IMPL_FILE_CHOOSER_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), XDP_IMPL_TYPE_FILE_CHOOSER_SKELETON, XdpImplFileChooserSkeletonClass))
#define XDP_IMPL_FILE_CHOOSER_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), XDP_IMPL_TYPE_FILE_CHOOSER_SKELETON, XdpImplFileChooserSkeletonClass))
#define XDP_IMPL_IS_FILE_CHOOSER_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), XDP_IMPL_TYPE_FILE_CHOOSER_SKELETON))
#define XDP_IMPL_IS_FILE_CHOOSER_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), XDP_IMPL_TYPE_FILE_CHOOSER_SKELETON))

typedef struct _XdpImplFileChooserSkeleton XdpImplFileChooserSkeleton;
typedef struct _XdpImplFileChooserSkeletonClass XdpImplFileChooserSkeletonClass;
typedef struct _XdpImplFileChooserSkeletonPrivate XdpImplFileChooserSkeletonPrivate;

struct _XdpImplFileChooserSkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  XdpImplFileChooserSkeletonPrivate *priv;
};

struct _XdpImplFileChooserSkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType xdp_impl_file_chooser_skeleton_get_type (void);

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (XdpImplFileChooserSkeleton, g_object_unref)
#endif

XdpImplFileChooser *xdp_impl_file_chooser_skeleton_new (void);


/* ------------------------------------------------------------------------ */
/* Declarations for org.freedesktop.impl.portal.Request */

#define XDP_IMPL_TYPE_REQUEST (xdp_impl_request_get_type ())
#define XDP_IMPL_REQUEST(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), XDP_IMPL_TYPE_REQUEST, XdpImplRequest))
#define XDP_IMPL_IS_REQUEST(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), XDP_IMPL_TYPE_REQUEST))
#define XDP_IMPL_REQUEST_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), XDP_IMPL_TYPE_REQUEST, XdpImplRequestIface))

struct _XdpImplRequest;
typedef struct _XdpImplRequest XdpImplRequest;
typedef struct _XdpImplRequestIface XdpImplRequestIface;

struct _XdpImplRequestIface
{
  GTypeInterface parent_iface;

  gboolean (*handle_close) (
    XdpImplRequest *object,
    GDBusMethodInvocation *invocation);

};

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (XdpImplRequest, g_object_unref)
#endif

GType xdp_impl_request_get_type (void);

GDBusInterfaceInfo *xdp_impl_request_interface_info (void);
guint xdp_impl_request_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void xdp_impl_request_complete_close (
    XdpImplRequest *object,
    GDBusMethodInvocation *invocation);



/* D-Bus method calls: */
void xdp_impl_request_call_close (
    XdpImplRequest *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean xdp_impl_request_call_close_finish (
    XdpImplRequest *proxy,
    GAsyncResult *res,
    GError **error);

gboolean xdp_impl_request_call_close_sync (
    XdpImplRequest *proxy,
    GCancellable *cancellable,
    GError **error);



/* ---- */

#define XDP_IMPL_TYPE_REQUEST_PROXY (xdp_impl_request_proxy_get_type ())
#define XDP_IMPL_REQUEST_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), XDP_IMPL_TYPE_REQUEST_PROXY, XdpImplRequestProxy))
#define XDP_IMPL_REQUEST_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), XDP_IMPL_TYPE_REQUEST_PROXY, XdpImplRequestProxyClass))
#define XDP_IMPL_REQUEST_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), XDP_IMPL_TYPE_REQUEST_PROXY, XdpImplRequestProxyClass))
#define XDP_IMPL_IS_REQUEST_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), XDP_IMPL_TYPE_REQUEST_PROXY))
#define XDP_IMPL_IS_REQUEST_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), XDP_IMPL_TYPE_REQUEST_PROXY))

typedef struct _XdpImplRequestProxy XdpImplRequestProxy;
typedef struct _XdpImplRequestProxyClass XdpImplRequestProxyClass;
typedef struct _XdpImplRequestProxyPrivate XdpImplRequestProxyPrivate;

struct _XdpImplRequestProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  XdpImplRequestProxyPrivate *priv;
};

struct _XdpImplRequestProxyClass
{
  GDBusProxyClass parent_class;
};

GType xdp_impl_request_proxy_get_type (void);

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (XdpImplRequestProxy, g_object_unref)
#endif

void xdp_impl_request_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
XdpImplRequest *xdp_impl_request_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
XdpImplRequest *xdp_impl_request_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void xdp_impl_request_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
XdpImplRequest *xdp_impl_request_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
XdpImplRequest *xdp_impl_request_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define XDP_IMPL_TYPE_REQUEST_SKELETON (xdp_impl_request_skeleton_get_type ())
#define XDP_IMPL_REQUEST_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), XDP_IMPL_TYPE_REQUEST_SKELETON, XdpImplRequestSkeleton))
#define XDP_IMPL_REQUEST_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), XDP_IMPL_TYPE_REQUEST_SKELETON, XdpImplRequestSkeletonClass))
#define XDP_IMPL_REQUEST_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), XDP_IMPL_TYPE_REQUEST_SKELETON, XdpImplRequestSkeletonClass))
#define XDP_IMPL_IS_REQUEST_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), XDP_IMPL_TYPE_REQUEST_SKELETON))
#define XDP_IMPL_IS_REQUEST_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), XDP_IMPL_TYPE_REQUEST_SKELETON))

typedef struct _XdpImplRequestSkeleton XdpImplRequestSkeleton;
typedef struct _XdpImplRequestSkeletonClass XdpImplRequestSkeletonClass;
typedef struct _XdpImplRequestSkeletonPrivate XdpImplRequestSkeletonPrivate;

struct _XdpImplRequestSkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  XdpImplRequestSkeletonPrivate *priv;
};

struct _XdpImplRequestSkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType xdp_impl_request_skeleton_get_type (void);

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (XdpImplRequestSkeleton, g_object_unref)
#endif

XdpImplRequest *xdp_impl_request_skeleton_new (void);


G_END_DECLS

#endif /* __XDG_DESKTOP_PORTAL_DBUS_H__ */
