
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "nautilus-enum-types.h"
/* Enumerations from "nautilus-enums.h" */
#include "/build/nautilus/src/build/../nautilus/src/nautilus-enums.h"
GType
nautilus_grid_icon_size_get_type (void)
{
    static GType type_once = 0;

    if (g_once_init_enter (&type_once))
    {
        static const GEnumValue values[] = {
            { NAUTILUS_GRID_ICON_SIZE_SMALL, "NAUTILUS_GRID_ICON_SIZE_SMALL", "small" },
            { NAUTILUS_GRID_ICON_SIZE_SMALL_PLUS, "NAUTILUS_GRID_ICON_SIZE_SMALL_PLUS", "small-plus" },
            { NAUTILUS_GRID_ICON_SIZE_MEDIUM, "NAUTILUS_GRID_ICON_SIZE_MEDIUM", "medium" },
            { NAUTILUS_GRID_ICON_SIZE_LARGE, "NAUTILUS_GRID_ICON_SIZE_LARGE", "large" },
            { NAUTILUS_GRID_ICON_SIZE_EXTRA_LARGE, "NAUTILUS_GRID_ICON_SIZE_EXTRA_LARGE", "extra-large" },
            { 0, NULL, NULL }
        };

        GType type = g_enum_register_static (g_intern_static_string ("NautilusGridIconSize"), values);

        g_once_init_leave (&type_once, type);
    }

    return type_once;
}
GType
nautilus_grid_zoom_level_get_type (void)
{
    static GType type_once = 0;

    if (g_once_init_enter (&type_once))
    {
        static const GEnumValue values[] = {
            { NAUTILUS_GRID_ZOOM_LEVEL_SMALL, "NAUTILUS_GRID_ZOOM_LEVEL_SMALL", "small" },
            { NAUTILUS_GRID_ZOOM_LEVEL_SMALL_PLUS, "NAUTILUS_GRID_ZOOM_LEVEL_SMALL_PLUS", "small-plus" },
            { NAUTILUS_GRID_ZOOM_LEVEL_MEDIUM, "NAUTILUS_GRID_ZOOM_LEVEL_MEDIUM", "medium" },
            { NAUTILUS_GRID_ZOOM_LEVEL_LARGE, "NAUTILUS_GRID_ZOOM_LEVEL_LARGE", "large" },
            { NAUTILUS_GRID_ZOOM_LEVEL_EXTRA_LARGE, "NAUTILUS_GRID_ZOOM_LEVEL_EXTRA_LARGE", "extra-large" },
            { 0, NULL, NULL }
        };

        GType type = g_enum_register_static (g_intern_static_string ("NautilusGridZoomLevel"), values);

        g_once_init_leave (&type_once, type);
    }

    return type_once;
}
GType
nautilus_list_icon_size_get_type (void)
{
    static GType type_once = 0;

    if (g_once_init_enter (&type_once))
    {
        static const GEnumValue values[] = {
            { NAUTILUS_LIST_ICON_SIZE_SMALL, "NAUTILUS_LIST_ICON_SIZE_SMALL", "small" },
            { NAUTILUS_LIST_ICON_SIZE_MEDIUM, "NAUTILUS_LIST_ICON_SIZE_MEDIUM", "medium" },
            { NAUTILUS_LIST_ICON_SIZE_LARGE, "NAUTILUS_LIST_ICON_SIZE_LARGE", "large" },
            { 0, NULL, NULL }
        };

        GType type = g_enum_register_static (g_intern_static_string ("NautilusListIconSize"), values);

        g_once_init_leave (&type_once, type);
    }

    return type_once;
}
GType
nautilus_list_zoom_level_get_type (void)
{
    static GType type_once = 0;

    if (g_once_init_enter (&type_once))
    {
        static const GEnumValue values[] = {
            { NAUTILUS_LIST_ZOOM_LEVEL_SMALL, "NAUTILUS_LIST_ZOOM_LEVEL_SMALL", "small" },
            { NAUTILUS_LIST_ZOOM_LEVEL_MEDIUM, "NAUTILUS_LIST_ZOOM_LEVEL_MEDIUM", "medium" },
            { NAUTILUS_LIST_ZOOM_LEVEL_LARGE, "NAUTILUS_LIST_ZOOM_LEVEL_LARGE", "large" },
            { 0, NULL, NULL }
        };

        GType type = g_enum_register_static (g_intern_static_string ("NautilusListZoomLevel"), values);

        g_once_init_leave (&type_once, type);
    }

    return type_once;
}
GType
nautilus_file_attributes_get_type (void)
{
    static GType type_once = 0;

    if (g_once_init_enter (&type_once))
    {
        static const GFlagsValue values[] = {
            { NAUTILUS_FILE_ATTRIBUTE_INFO, "NAUTILUS_FILE_ATTRIBUTE_INFO", "info" },
            { NAUTILUS_FILE_ATTRIBUTE_DEEP_COUNTS, "NAUTILUS_FILE_ATTRIBUTE_DEEP_COUNTS", "deep-counts" },
            { NAUTILUS_FILE_ATTRIBUTE_DIRECTORY_ITEM_COUNT, "NAUTILUS_FILE_ATTRIBUTE_DIRECTORY_ITEM_COUNT", "directory-item-count" },
            { NAUTILUS_FILE_ATTRIBUTE_THUMBNAIL_INFO, "NAUTILUS_FILE_ATTRIBUTE_THUMBNAIL_INFO", "thumbnail-info" },
            { NAUTILUS_FILE_ATTRIBUTE_EXTENSION_INFO, "NAUTILUS_FILE_ATTRIBUTE_EXTENSION_INFO", "extension-info" },
            { NAUTILUS_FILE_ATTRIBUTE_THUMBNAIL_BUFFER, "NAUTILUS_FILE_ATTRIBUTE_THUMBNAIL_BUFFER", "thumbnail-buffer" },
            { NAUTILUS_FILE_ATTRIBUTE_MOUNT, "NAUTILUS_FILE_ATTRIBUTE_MOUNT", "mount" },
            { NAUTILUS_FILE_ATTRIBUTE_FILESYSTEM_INFO, "NAUTILUS_FILE_ATTRIBUTE_FILESYSTEM_INFO", "filesystem-info" },
            { 0, NULL, NULL }
        };

        GType type = g_flags_register_static (g_intern_static_string ("NautilusFileAttributes"), values);

        g_once_init_leave (&type_once, type);
    }

    return type_once;
}
GType
nautilus_mode_get_type (void)
{
    static GType type_once = 0;

    if (g_once_init_enter (&type_once))
    {
        static const GEnumValue values[] = {
            { NAUTILUS_MODE_BROWSE, "NAUTILUS_MODE_BROWSE", "browse" },
            { NAUTILUS_MODE_OPEN_FILE, "NAUTILUS_MODE_OPEN_FILE", "open-file" },
            { NAUTILUS_MODE_OPEN_FILES, "NAUTILUS_MODE_OPEN_FILES", "open-files" },
            { NAUTILUS_MODE_OPEN_FOLDER, "NAUTILUS_MODE_OPEN_FOLDER", "open-folder" },
            { NAUTILUS_MODE_OPEN_FOLDERS, "NAUTILUS_MODE_OPEN_FOLDERS", "open-folders" },
            { NAUTILUS_MODE_SAVE_FILE, "NAUTILUS_MODE_SAVE_FILE", "save-file" },
            { NAUTILUS_MODE_SAVE_FILES, "NAUTILUS_MODE_SAVE_FILES", "save-files" },
            { 0, NULL, NULL }
        };

        GType type = g_enum_register_static (g_intern_static_string ("NautilusMode"), values);

        g_once_init_leave (&type_once, type);
    }

    return type_once;
}
GType
nautilus_open_flags_get_type (void)
{
    static GType type_once = 0;

    if (g_once_init_enter (&type_once))
    {
        static const GFlagsValue values[] = {
            { NAUTILUS_OPEN_FLAG_NORMAL, "NAUTILUS_OPEN_FLAG_NORMAL", "normal" },
            { NAUTILUS_OPEN_FLAG_NEW_WINDOW, "NAUTILUS_OPEN_FLAG_NEW_WINDOW", "new-window" },
            { NAUTILUS_OPEN_FLAG_NEW_TAB, "NAUTILUS_OPEN_FLAG_NEW_TAB", "new-tab" },
            { NAUTILUS_OPEN_FLAG_DONT_MAKE_ACTIVE, "NAUTILUS_OPEN_FLAG_DONT_MAKE_ACTIVE", "dont-make-active" },
            { NAUTILUS_OPEN_FLAG_REUSE_EXISTING, "NAUTILUS_OPEN_FLAG_REUSE_EXISTING", "reuse-existing" },
            { 0, NULL, NULL }
        };

        GType type = g_flags_register_static (g_intern_static_string ("NautilusOpenFlags"), values);

        g_once_init_leave (&type_once, type);
    }

    return type_once;
}
GType
nautilus_search_time_type_get_type (void)
{
    static GType type_once = 0;

    if (g_once_init_enter (&type_once))
    {
        static const GEnumValue values[] = {
            { NAUTILUS_SEARCH_TIME_TYPE_LAST_MODIFIED, "NAUTILUS_SEARCH_TIME_TYPE_LAST_MODIFIED", "last-modified" },
            { NAUTILUS_SEARCH_TIME_TYPE_LAST_ACCESS, "NAUTILUS_SEARCH_TIME_TYPE_LAST_ACCESS", "last-access" },
            { NAUTILUS_SEARCH_TIME_TYPE_CREATED, "NAUTILUS_SEARCH_TIME_TYPE_CREATED", "created" },
            { 0, NULL, NULL }
        };

        GType type = g_enum_register_static (g_intern_static_string ("NautilusSearchTimeType"), values);

        g_once_init_leave (&type_once, type);
    }

    return type_once;
}
GType
nautilus_sidebar_row_type_get_type (void)
{
    static GType type_once = 0;

    if (g_once_init_enter (&type_once))
    {
        static const GEnumValue values[] = {
            { NAUTILUS_SIDEBAR_ROW_INVALID, "NAUTILUS_SIDEBAR_ROW_INVALID", "row-invalid" },
            { NAUTILUS_SIDEBAR_ROW_BUILT_IN, "NAUTILUS_SIDEBAR_ROW_BUILT_IN", "row-built-in" },
            { NAUTILUS_SIDEBAR_ROW_EXTERNAL_MOUNT, "NAUTILUS_SIDEBAR_ROW_EXTERNAL_MOUNT", "row-external-mount" },
            { NAUTILUS_SIDEBAR_ROW_INTERNAL_MOUNT, "NAUTILUS_SIDEBAR_ROW_INTERNAL_MOUNT", "row-internal-mount" },
            { NAUTILUS_SIDEBAR_ROW_NEW_BOOKMARK, "NAUTILUS_SIDEBAR_ROW_NEW_BOOKMARK", "row-new-bookmark" },
            { NAUTILUS_SIDEBAR_ROW_BOOKMARK, "NAUTILUS_SIDEBAR_ROW_BOOKMARK", "row-bookmark" },
            { NAUTILUS_SIDEBAR_ROW_BOOKMARK_PLACEHOLDER, "NAUTILUS_SIDEBAR_ROW_BOOKMARK_PLACEHOLDER", "row-bookmark-placeholder" },
            { NAUTILUS_SIDEBAR_N_ROW_TYPES, "NAUTILUS_SIDEBAR_N_ROW_TYPES", "n-row-types" },
            { 0, NULL, NULL }
        };

        GType type = g_enum_register_static (g_intern_static_string ("NautilusSidebarRowType"), values);

        g_once_init_leave (&type_once, type);
    }

    return type_once;
}
GType
nautilus_sidebar_section_type_get_type (void)
{
    static GType type_once = 0;

    if (g_once_init_enter (&type_once))
    {
        static const GEnumValue values[] = {
            { NAUTILUS_SIDEBAR_SECTION_INVALID, "NAUTILUS_SIDEBAR_SECTION_INVALID", "section-invalid" },
            { NAUTILUS_SIDEBAR_SECTION_DEFAULT_LOCATIONS, "NAUTILUS_SIDEBAR_SECTION_DEFAULT_LOCATIONS", "section-default-locations" },
            { NAUTILUS_SIDEBAR_SECTION_BOOKMARKS, "NAUTILUS_SIDEBAR_SECTION_BOOKMARKS", "section-bookmarks" },
            { NAUTILUS_SIDEBAR_SECTION_CLOUD, "NAUTILUS_SIDEBAR_SECTION_CLOUD", "section-cloud" },
            { NAUTILUS_SIDEBAR_SECTION_MOUNTS, "NAUTILUS_SIDEBAR_SECTION_MOUNTS", "section-mounts" },
            { NAUTILUS_SIDEBAR_N_SECTION_TYPES, "NAUTILUS_SIDEBAR_N_SECTION_TYPES", "n-section-types" },
            { 0, NULL, NULL }
        };

        GType type = g_enum_register_static (g_intern_static_string ("NautilusSidebarSectionType"), values);

        g_once_init_leave (&type_once, type);
    }

    return type_once;
}
/* Enumerations from "nautilus-search-popover.h" */
#include "/build/nautilus/src/build/../nautilus/src/nautilus-search-popover.h"
GType
nautilus_search_filter_get_type (void)
{
    static GType type_once = 0;

    if (g_once_init_enter (&type_once))
    {
        static const GEnumValue values[] = {
            { NAUTILUS_SEARCH_FILTER_CONTENT, "NAUTILUS_SEARCH_FILTER_CONTENT", "content" },
            { NAUTILUS_SEARCH_FILTER_DATE, "NAUTILUS_SEARCH_FILTER_DATE", "date" },
            { NAUTILUS_SEARCH_FILTER_LAST, "NAUTILUS_SEARCH_FILTER_LAST", "last" },
            { NAUTILUS_SEARCH_FILTER_TYPE, "NAUTILUS_SEARCH_FILTER_TYPE", "type" },
            { 0, NULL, NULL }
        };

        GType type = g_enum_register_static (g_intern_static_string ("NautilusSearchFilter"), values);

        g_once_init_leave (&type_once, type);
    }

    return type_once;
}

/* Generated data ends here */

